/**************************************************************************
 *
 * FILE  g64scan.c
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: g64scan.c,v 1.33 2007-09-24 09:30:05 tlr Exp $
 *
 * DESCRIPTION
 *   This program scans a g64 file and presents the results to the screen.
 *
 * HISTORY
 *   0.3 (20070325) tlr 
 *     - handles short data blocks more gently.
 *
 *   0.2 (20070325) tlr 
 *     - handles short data blocks more gently.
 *     - handles non sync aligned circular tracks.
 *
 *   0.1 (20070302) tlr 
 *     - first version.
 *
 * LICENSE
 *   g64tools - C64 disk image tools
 *
 *   Copyright (c) 2007 Daniel Kahlin <daniel@kahlin.net>
 *   All rights reserved.
 * 
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither the names of the copyright holders nor the names of their
 *      contributors may be used to endorse or promote products derived 
 *      from this software without specific prior written permission.
 * 
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   ``AS IS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
 *   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include "global.h"
#include "util.h"
#include "1541.h"
#include "gcrimage.h"

#define PROGRAM "g64scan"
#define VERSION "0.4"

const char program_g[]=PROGRAM;

float track_g;

void scan_disk(struct Image *img, int half_tracks);
int dump_track(struct Track *trk);

void print_bufb(int *buf, int start, int len);
void print_buf(struct Track *trk, int sb, int eb);
void print_buf_gcr(int *buf, int len, int bit, int trklenbits);

int main(int argc, char *argv[])
{
    int c;
    char *name="";

    /* defaults */
    debug_g=0;
    verbose_g=0;
    sync_threshold_g=12;
    track_g=-1;

    /*
     * scan for valid options
     */
    while (EOF!=(c=getopt (argc, argv, "t:S:dvVh"))) {
        switch (c) {
	
	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit (1);

	/* print version */
	case 'v':
	    verbose_g=1;
	    break;
	/* print version */
	case 'd':
	    debug_g=1;
	    break;

	/* print version */
	case 'V':
	    fprintf (stdout, PROGRAM " " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " " VERSION ": C64 disk image analyzer\n"
"Copyright (c) 2007 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-t<track>][-S<threshold>[-d][-v][-h][-V] <file>\n"
"\n"
"Valid options:\n"
"    -t              track (if specified, dump details of track)\n"
"    -S              set sync threshold (default: 12)\n"
"    -d              debugging output\n"
"    -v              be verbose\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
"\n"
	    );
	    exit(0);
 
	/* set sync threshold */
	case 't':
	    track_g=atof(optarg);
	    break;

	/* set sync threshold */
	case 'S':
	    sync_threshold_g=atoi(optarg);
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     * we expect one more argument (inname)
     */
    if (argc-optind < 1)
	panic("too few arguments");

    name=argv[optind];

    /* 
     * open files
     */
    struct Image *img=open_image(name);

    if (track_g>0) {
	int ind=track2index(track_g);
	struct Track *trk=read_track(img, ind);
	dump_track(trk);
	free(trk);
    } else
	scan_disk(img, img->head.num_halftracks);

    close_image(img);
    exit(0);
}


/**************************************************************************
 *
 * NAME  scan_disk
 *
 * DESCRIPTION
 *   Scan a full disk image, and display as a matrix.
 *
 ******/
void scan_disk(struct Image *img, int half_tracks)
{
    int i,j,k;
    int numtr=img->head.num_tracks;
    int incr=(half_tracks)?1:2;
    int scan_res[MAX_TRACKS][MAX_SECTORS];
    uint8_t buf[256];
    uint8_t id[2];

    /* try to read 18,0 to obtain the disk id */
    struct Track *trk=read_track(img, track2index(18.0));
    read_sector(trk,0,buf,id);
    free(trk);
    printf("Disk ID='%c%c'\n",id[0],id[1]);

    /* scan for sectors from the image */
    for (i=0; i<numtr; i+=incr) {
	struct Track *trk=read_track(img, i);
	for (j=0; j<num_sectors(i); j++) {
	    scan_res[i][j]=scan_sector(trk,j, id);
	}
	free(trk);
    }

    /* present information to the screen */
    printf("    ");
    for (i=0; i<numtr; i+=incr) {
	if (i%20 == 0)
	    printf("%d",(i+2)/20);
	else if (i == (18-1)*2)
	    printf("1");
	else
	    printf(" ");
    }
    printf("\n    ");
    for (i=0; i<numtr; i+=incr) {
	if (i%2 == 0)
	    printf("%d",((i+2)/2) % 10 );
	else
	    printf("|");
    }
    printf("\n");
    printf("   +");
    for (i=0; i<numtr; i+=incr) {
	printf("-");
    }
    printf("+\n");
    int last=numtr;
    int diff=0;
    for (j=0; j<21; j++) {
	printf("%2d |",j);
	for (i=0; i<numtr; i+=incr) {
	    if (j>=num_sectors(i))
		break;
	    char errchars[]="?.2s456789ABCDEF";
	    printf("%c",errchars[scan_res[i][j]]);   
	}
	diff=last-i;
	last=i;
	if (!diff) {
	    printf("|\n");
	} else {
	    printf("+");
	    for (k=0; k<diff-incr; k+=incr)
		printf("-");
	    printf("+\n");
	}
    }
    diff=last;
    printf("   +");
    for (k=0; k<diff; k+=incr) {
	printf("-");
    }
    printf("+\n");
}
    
/**************************************************************************
 *
 * NAME  dump_track
 *
 * DESCRIPTION
 *   Dump contents of a track as GCR and decoded data.
 *
 ******/
int dump_track(struct Track *trk)
{
    int i;
    int sb, eb;
    int decode_buf[MAX_DECODELEN];
    int decode_len;

    for (i=0; i < trk->num_syncs; i++) {
	sb=trk->sync_start[i];
	eb=trk->sync_end[i];
	sb -= (8 - ((eb-sb) % 8)) % 8;
	printf("--- sync -----------------------\n");
	print_buf(trk, sb, eb);

	sb=trk->sync_end[i];
	eb=trk->sync_start[i+1];
	printf("--- gcr ------------------------\n");
	print_buf(trk, sb, eb);
	printf("--- decoded --------------------\n");
	decode_len=gcr_decode_block(trk, sb, eb, decode_buf);
	print_buf_gcr(decode_buf, decode_len, sb, trk->track_len_bits);
	printf("--- interpretation -------------\n");
	switch (decode_buf[0]) {
	case 0x08:
	    {
		uint8_t chk=decode_buf[1];
		uint8_t s=decode_buf[2];
		uint8_t t=decode_buf[3];
		uint8_t id2=decode_buf[4];
		uint8_t id1=decode_buf[5];
		uint8_t calc=s^t^id2^id1;
		printf(
                    "sector header, t=%d, s=%d (ID='%c%c') chk=$%02x (calc=$%02x)\n",
		    t,s,id1,id2,chk,calc
		);
		break;
	    }
	case 0x07:
	    {
		if (decode_len >= 258) {
		    int i;
		    int chk;
		    uint8_t calc=0;
		    chk=decode_buf[257];
		    for (i=0; i<256; i++)
			calc^=decode_buf[i+1];
		    printf(
		        "sector data, chk=$%02x (calc=$%02x)\n",
			chk,calc
		    );
		    print_bufb(decode_buf+1, 0, 256);
		} else {
		    printf(
		        "sector data, short sector!\n"
		    );
		    if (decode_len > 1)
			print_bufb(decode_buf+1, 0, decode_len-1);
		    else
			printf("no data!\n");
		}
		break;
	    }
	default:
	    printf("unknown\n");
	}
	printf("--------------------------------\n\n");
    }
    return 0;
}


void print_bufb(int *buf, int start, int len)
{
    int i,c=0;
    int d;
    for (i=0; i<len; i++) {
	d=buf[start+i];
	if (!(c%16)) {
	    printf("%04x: ",i);
	}
	if (d < 0) {
	    printf("xx ");
	} else {
	    printf("%02x ",d);
	}
	if ((c%16)==15) {
	    printf("\n");
	}
	c++;
    }
    if ((c%16)) {
	printf("\n");
    }
}

void print_buf(struct Track *trk, int sb, int eb)
{
    int i,c=0;
    uint8_t d;
    if (eb<sb)
	eb+=trk->track_len_bits;

    for (i=sb; i<eb-7; i+=8) {
	d=get_byte(trk, i);
	if (!(c%24)) {
	    printf("%04x: ",i >> 3);
	}
	printf("%02x ",d);
	if ((c%24)==23) {
	    printf("\n");
	}
	c++;
    }
    if ((c%24)) {
	printf("\n");
    }
}

void print_buf_gcr(int *buf, int len, int bit, int trklenbits)
{
    int i,c=0;
    int decode;
    for (i=0; i<len; i++) {
	if (!(c%16)) {
	    printf("%04x: ",bit >> 3);
	}
	decode=buf[i];
	if (decode < 0) {
	    printf("xx ");
	} else {
	    printf("%02x ",decode);
	}
	if ((c%16)==15) {
	    printf("\n");
	}
	bit+=10;
	bit%=trklenbits;
	c++;
    }
    if ((c%16)) {
	printf("\n");
    }
}

/* eof */
