/**************************************************************************
 *
 * FILE  gcrimage.c
 * Copyright (c) 2007 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: gcrimage.c,v 1.1 2007-03-31 08:34:50 tlr Exp $
 *
 * DESCRIPTION
 *   This program scans a g64 file and presents the results to the screen.
 *
 ******/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "global.h"
#include "util.h"
#include "1541.h"
#include "gcrimage.h"

int sync_threshold_g;

/**************************************************************************
 *
 * NAME  scan_sector
 *
 * DESCRIPTION
 *   Scan for one particular sector, and try to return correct FDC error
 *   codes.
 *
 ******/
int scan_sector(struct Track *trk, int sector, uint8_t *id)
{
    int i;
    enum { SEARCH, FOUND } state = SEARCH;
    int sb, eb;
    int decode_len;
    int decode_buf[MAX_DECODELEN];

    if (!trk->num_syncs)
	return 3; /* No SYNC sequence found */
    for (i=0; i < trk->num_syncs+1; i++) {
	sb=trk->sync_end[i % trk->num_syncs];
	eb=trk->sync_start[(i+1) % trk->num_syncs];
	decode_len=gcr_decode_block(trk, sb, eb, decode_buf);
	switch (decode_buf[0]) {
	case 0x08: /* header block */
	    {
		if (state == FOUND)
		    return 4; /* data descriptor byte not found */
		uint8_t chk=decode_buf[1];
		uint8_t s=decode_buf[2];
		uint8_t t=decode_buf[3];
		uint8_t id2=decode_buf[4];
		uint8_t id1=decode_buf[5];
		uint8_t calc=s^t^id2^id1;
		if (chk != calc)
		    return 9; /* Checksum error in header block */
		if (id) {
		    if (id1 != id[0] || id2 != id[1])
			return 11; /* Disk sector ID mismatch */
		}
		if (sector == s)
		    state = FOUND;
		break;
	    }
	case 0x07: /* data block */
	    {
		if (state != FOUND)
		    break;
		if (decode_len >= 258) {
		    int i;
		    uint8_t chk=decode_buf[257];
		    uint8_t calc=0;
		    for (i=0; i<256; i++)
			calc^=decode_buf[i+1];
		    if (calc != chk) 
			return 5; /* checksum error in data block */
		    return 1; /* ok */
		} else
		    return 5; /* short data block */
	    }
	default:
	    break;
	}
    }
    if (state == FOUND)
	return 4; /* data descriptor byte not found */
    return 2; /* header block not found */
}

/**************************************************************************
 *
 * NAME  read_sector
 *
 * DESCRIPTION
 *   Read sector, and try to return correct FDC error
 *   codes.
 *
 ******/
int read_sector(struct Track *trk, int sector, uint8_t *buf, uint8_t *id)
{
    int i;
    enum { SEARCH, FOUND } state = SEARCH;
    int sb, eb;
    int decode_len;
    int decode_buf[MAX_DECODELEN];

    if (!trk->num_syncs)
	return 3; /* No SYNC sequence found */
    for (i=0; i < trk->num_syncs+1; i++) {
	sb=trk->sync_end[i % trk->num_syncs];
	eb=trk->sync_start[(i+1) % trk->num_syncs];
	decode_len=gcr_decode_block(trk, sb, eb, decode_buf);
	switch (decode_buf[0]) {
	case 0x08: /* header block */
	    {
		if (state == FOUND)
		    return 4; /* data descriptor byte not found */
		uint8_t chk=decode_buf[1];
		uint8_t s=decode_buf[2];
		uint8_t t=decode_buf[3];
		uint8_t id2=decode_buf[4];
		uint8_t id1=decode_buf[5];
		uint8_t calc=s^t^id2^id1;
		if (chk != calc)
		    return 9; /* Checksum error in header block */
		if (id) {
		    id[0]=id1;
		    id[1]=id2;
		}
		if (sector == s)
		    state = FOUND;
		break;
	    }
	case 0x07: /* data block */
	    {
		if (state != FOUND)
		    break;
		if (decode_len >= 258) {
		    int i;
		    uint8_t chk=decode_buf[257];
		    uint8_t calc=0;
		    for (i=0; i<256; i++) {
			buf[i]=decode_buf[i+1];
			calc^=decode_buf[i+1];
		    }
		    if (calc != chk) 
			return 5; /* checksum error in data block */
		    return 1; /* ok */
		} else
		    return 5; /* short data block */
	    }
	default:
	    break;
	}
    }
    if (state == FOUND)
	return 4; /* data descriptor byte not found */
    return 2; /* header block not found */
}



/**************************************************************************
 *
 * NAME  open_image
 *
 * DESCRIPTION
 *   Open disk image and return a handle.
 *
 ******/
struct Image *open_image(const char *name)
{
    /* working variables */
    int i;

    /* header info */
    char buf[16];
    uint8_t version;
    struct Image *img;

    img=malloc(sizeof(struct Image));
    if (!img)
	panic("couldn't malloc image");
    img->fp=fopen(name,"rb");
    if (!img->fp)
	panic("couldn't open file for reading");

    /* read header data */
    img->head.type = -1;
    fread(buf,1,16,img->fp);
    if (!strncmp(buf, "GCR-1541", 8)) {
	img->head.type = TYPE_G64;
    }
    if (!strncmp(buf, "MNIB-1541-RAW", 13)) {
	img->head.type = TYPE_NIB;
	panic("nib/nb2 file not yet supported");
    }
    if (img->head.type < 0)
	panic("not a supported file format");

    fseek(img->fp, 8, SEEK_SET);
    version=fgetc(img->fp);
    img->head.num_tracks=fgetc(img->fp);
    img->head.track_size=fgetc(img->fp);
    img->head.track_size|=fgetc(img->fp) << 8;
    if (feof(img->fp))
	panic("unexpected end of file");
    for (i=0; i < img->head.num_tracks; i++) {
	uint32_t offs;
	offs = fgetc(img->fp);
	offs |= fgetc(img->fp) << 8;
	offs |= fgetc(img->fp) << 16;
	offs |= fgetc(img->fp) << 24;
	img->head.track_offs[i]=offs;
    }
    for (i=0; i < img->head.num_tracks; i++) {
	uint32_t offs;
	offs = fgetc(img->fp);
	offs |= fgetc(img->fp) << 8;
	offs |= fgetc(img->fp) << 16;
	offs |= fgetc(img->fp) << 24;
	img->head.speed_offs[i]=offs;
    }
    if (feof(img->fp))
	panic("unexpected end of file");
    /* header read! */
    printf(
	"g64 version=%d, %d track entries of max %d bytes each.\n",
	version,
	img->head.num_tracks,
	img->head.track_size
    );

    /* check for highest used entry */
    for (i=img->head.num_tracks-1; i!=0; i--)
	if (img->head.track_offs[i])
	    break;
    img->head.track_max = i;

    /* check for half tracks */
    img->head.num_halftracks=0;
    for (i=1; i<=img->head.track_max; i+=2) {
	if (img->head.track_offs[i])
	    img->head.num_halftracks++;
    }

    /* check for speed offset blocks */
    img->head.num_speedblocks=0;
    for (i=0; i<=img->head.track_max; i++) {
	if (img->head.speed_offs[i] > 4)
	    img->head.num_speedblocks++;
    }
    printf(
        "highest track %.1f, %d half tracks, %d speed offset blocks.\n",
	index2track(img->head.track_max),
	img->head.num_halftracks,
	img->head.num_speedblocks
    );
    return img;
}

/**************************************************************************
 *
 * NAME  close_image
 *
 * DESCRIPTION
 *   Close a disk image previously returned by open_image.
 *
 ******/
void close_image(struct Image *img)
{
    fclose(img->fp);
    free(img);
}

/**************************************************************************
 *
 * NAME  read_track
 *
 * DESCRIPTION
 *   Read track from image.
 *
 ******/
struct Track *read_track(struct Image *img, int track) 
{
    /* working variables */
    int i;
    int bit, sb, eb, cnt;

    struct Header *head=&(img->head);
    /* track data */
    struct Track *trk;
    trk=malloc(sizeof(struct Track));
    if (!trk)
	panic("couldn't malloc track");

    trk->track=track;
    /* read track data */
    fseek(img->fp, head->track_offs[track], SEEK_SET);
    trk->track_len = fgetc(img->fp);
    trk->track_len |= fgetc(img->fp) << 8;
    if (feof(img->fp))
	panic("unexpected end of file");
    fread(trk->trackbuf, 1, trk->track_len, img->fp);
    if (feof(img->fp))
	panic("unexpected end of file");

    if (head->speed_offs[track] > 4) {
	fseek(img->fp, head->speed_offs[track], SEEK_SET);
	for (i=0; i < trk->track_len+3; i+=4) {
	    uint8_t tmp;
	    tmp=fgetc(img->fp);
	    trk->speedbuf[i+0]=(tmp >> 6) & 0x03;
	    trk->speedbuf[i+1]=(tmp >> 4) & 0x03;
	    trk->speedbuf[i+2]=(tmp >> 2) & 0x03;
	    trk->speedbuf[i+3]=(tmp >> 0) & 0x03;
	}
	if (feof(img->fp))
	    panic("unexpected end of file");
    } else {
	for (i=0; i<trk->track_len; i++)
	    trk->speedbuf[i]=head->speed_offs[track];
    }
    if (debug_g) printf(
	"track: %.1f, %d bytes raw\n",
	index2track(track),
	trk->track_len
    );
    /* scan for sync marks =sync_threshold_g or more '1's */
    trk->track_len_bits=trk->track_len*8;
    sb=eb=bit=0;
    cnt=0;
    trk->num_syncs=0;
    while (bit < trk->track_len_bits) {
	if (get_bit(trk, bit)) {
	    if (cnt==0)
		sb=bit;
	    cnt++;
	} else {
	    if (cnt > sync_threshold_g) {
		eb=bit;
		trk->sync_start[trk->num_syncs]=sb;
		trk->sync_end[trk->num_syncs]=eb;
		if (debug_g) printf("sync: %d bits @ %d to %d\n", cnt, sb, eb);
		trk->num_syncs++;
	    }
	    cnt=0;
	}
	bit++;
    }
    /* fake entry to simplify parsing */
    trk->sync_start[trk->num_syncs]=trk->sync_start[0];
    /* syncs found, present data */
    return trk;
}

uint8_t get_bit(struct Track *trk, int bit)
{
    int o=bit >> 3;
    int b=7-(bit & 0x7);
    uint8_t *buf=trk->trackbuf;
    int trklen=trk->track_len;
    return (buf[o%trklen]>>b) & 0x01;   
}
uint8_t get_byte(struct Track *trk, int bit)
{
    int o=bit >> 3;
    int b=(bit & 0x7);
    uint8_t *buf=trk->trackbuf;
    int trklen=trk->track_len;
    return (buf[o%trklen]<<b) | (buf[(o+1)%trklen]>>(8-b));   
}
uint16_t get_10bit(struct Track *trk, int bit)
{
    uint8_t b1=get_byte(trk,bit);
    uint8_t b2=get_byte(trk,bit+8);
    return (b1 << 2) | (b2 >> 6);
}

int gcr_tab[] = {
     -1, -1, -1, -1, -1, -1, -1, -1,
     -1,0x8,0x0,0x1, -1,0xc,0x4,0x5,
     -1, -1,0x2,0x3, -1,0xf,0x6,0x7,
     -1,0x9,0xa,0xb, -1,0xd,0xe, -1
};

int gcr_decode(uint16_t d)
{
    uint8_t n1=(d>>5) & 0x1f;
    uint8_t n2=d & 0x1f;
    int d1=gcr_tab[n1];
    int d2=gcr_tab[n2];
    if (d1<0 || d2<0)
	return -1;
    return (d1 << 4) | d2;
}


int gcr_decode_block(struct Track *trk, int sb, int eb, int *dest)
{
    int i, j;
    uint16_t d;
    if (eb<sb)
	eb+=trk->track_len_bits;

    for (j=0, i=sb; i<eb-9; i+=10) {
	d=get_10bit(trk, i);
	dest[j]=gcr_decode(d);
	j++;
    }
    return j;
}

/* eof */
